#ifndef DISPred_DISPredictor_h
#define DISPred_DISPredictor_h

#include "ControlCards.h"
#include "RedSigmaGrid.h"
#include "DiffSigmaPoint.h"

#include <sstream>
#include <string>
#include <fstream>
#include "TString.h"
#include "TFile.h"
#include "TTree.h"
#include "TMath.h"
#include "TGraphAsymmErrors.h"
#include "TH1D.h"
#include "qcdnum.h"
#include <iostream>
#include <math.h>
// need LHAPDF
#include "LHAPDF/LHAPDF.h"

// GSL for integration
#include <gsl/gsl_math.h>
#include <gsl/gsl_monte.h>
#include <gsl/gsl_monte_plain.h>
#include <gsl/gsl_monte_miser.h>
#include <gsl/gsl_monte_vegas.h>


namespace DISPred  {
  class DISPredictor {
    
  public: 

    static DISPredictor* Instance();

    void Initialise(int my_argc, char **my_argv);
    void CalculateCrossSections();    
    void InitPDF(int subset);
    void PrintResults();
    void WriteOutput();

    double CalculateReducedCrossSection(double x, double q2);
    double CalculateCCReducedCrossSection(double x, double q2);
    double CalculatePropagator(double q2, double x);
    double CalculateCCPropagator(double q2, double x);
    double CalculateDSigmaDQ2(double q2min,double q2max);
    double CalculateDSigmaDX(double xmin,double xmax);
    double CalculateQ2DSigmaDQ2(double q2min,double q2max);
    double CalculateXDSigmaDX(double xmin,double xmax);
    double CalculateDSigmaDY(double ymin,double ymax);
    double CalculateYDSigmaDY(double ymin,double ymax);
    double S(){return _s;};

  private:


    TFile *my_file;
    TTree *out_tree;
    TH1D * MakeSingleDiffHisto( std::string fileName , std::string hist_name, double *binedges);
    void InitSingleDifferentials();
    void CalculateCentreOfMassEnergy();
    double FL(double x, double q2);
    double FLP(double x, double q2);
    double FL_LO(double x, double q2);
    double FL_NLO(double x, double q2);
    double FLP_LO(double x, double q2);
    double FLP_NLO(double x, double q2);
    double F2(double x, double q2);
    double F2P(double x, double q2);
    double F2_LO(double x, double q2);
    double F2_NLO(double x, double q2);
    double F2P_LO(double x, double q2);
    double F2P_NLO(double x, double q2);
    double XF3(double x, double q2);
    double XF3P(double x, double q2);
    double XF3_LO(double x, double q2);
    double XF3_NLO(double x, double q2);
    double XF3P_LO(double x, double q2);
    double XF3P_NLO(double x, double q2);
    double FLCC(double x, double q2);
    double FLCC_LO(double x, double q2);
    double FLCC_NLO(double x, double q2);
    double F2CC(double x, double q2);
    double F2CC_LO(double x, double q2);
    double F2CC_NLO(double x, double q2);
    double XF3CC(double x, double q2);
    double XF3CC_LO(double x, double q2);
    double XF3CC_NLO(double x, double q2);
    void InitCards();
    void ReadCards(int my_argc, char **my_argv);
    int ReadRedSigmaGrid();
    int ReadDiffSigmaPoints(TString cardfile, std::vector <DISPred::DiffSigmaPoint> *points );
    double SigmaInteg(double xmin, double xmax, double q2min, double q2max);
    double SigmaIntegY(double ymin, double ymax, double q2min, double q2max);
    double YSigmaIntegY(double ymin, double ymax, double q2min, double q2max);
    double Q2SigmaInteg(double xmin, double xmax, double q2min, double q2max);
    double XSigmaInteg(double xmin, double xmax, double q2min, double q2max);

    double _s;
    int _Tpoint;  // variable for tree
    double _Tq2; // variable for tree
    double _Tx; // variable for tree
    double _Td; // variable for tree
    double _Tu; // variable for tree
    double _Ts; // variable for tree
    double _Tc; // variable for tree
    double _Tb; // variable for tree
    double _Tdbar; // variable for tree
    double _Tubar; // variable for tree
    double _Tsbar; // variable for tree
    double _Tcbar; // variable for tree
    double _Tbbar; // variable for tree
    double _Tredsigma; // variable for tree
    double _Tddiffsigma; // variable for tree

    RedSigmaGrid *_redSigmas;
    TH1D  *_dSigdX;
    TH1D  *_dSigdQ2;
    TH1D  *_dSigdY;
    TGraphAsymmErrors  *_GdSigdX;
    TGraphAsymmErrors  *_GdSigdQ2;
    TGraphAsymmErrors  *_GdSigdY;
    TH1D  *_XdSigdX;
    TH1D  *_Q2dSigdQ2;
    TH1D  *_YdSigdY;
    double _q2binedges[10000]; // can be increased if necessary
    double _xbinedges[10000];
    double _ybinedges[10000];
    float _q2points[10000]; // can be increased if necessary
    float _xpoints[10000];
    float _ypoints[10000];
    float _q2vals[10000]; // can be increased if necessary
    float _xvals[10000];
    float _yvals[10000];
    float _q2errs[10000]; // can be increased if necessary
    float _xerrs[10000];
    float _yerrs[10000];
    std::vector <DISPred::DiffSigmaPoint> * _dSigmaDQ2;
    std::vector <DISPred::DiffSigmaPoint> * _dSigmaDX;
    std::vector <DISPred::DiffSigmaPoint> * _dSigmaDY;
  };
}


double g (double *k, size_t dim, void *params)
{
  if (k[1]<DISPred::ControlCards::Instance()->fetchValueDouble("XMax") && k[1]>DISPred::ControlCards::Instance()->fetchValueDouble("XMin") && k[0]<DISPred::ControlCards::Instance()->fetchValueDouble("Q2Max") && k[0]>DISPred::ControlCards::Instance()->fetchValueDouble("Q2Min") && DISPred::ControlCards::Instance()->fetchValueDouble("YMax") > k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) &&  DISPred::ControlCards::Instance()->fetchValueDouble("YMin") <k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])  ){  
    if (!strcmp(DISPred::ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){  
    return DISPred::DISPredictor::Instance()->CalculateCCReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculateCCPropagator(k[0],k[1] );
    } else {
      return DISPred::DISPredictor::Instance()->CalculateReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculatePropagator(k[0],k[1] );
    }

  } else{
    return 0.0;
  }
}


double Yg (double *k, size_t dim, void *params)
{
  double jacob=k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]*k[1]); // multiply by (-q2)/(s*y**2)
if (k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])<DISPred::ControlCards::Instance()->fetchValueDouble("XMax") && k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])>DISPred::ControlCards::Instance()->fetchValueDouble("XMin") && k[0]<DISPred::ControlCards::Instance()->fetchValueDouble("Q2Max") && k[0]>DISPred::ControlCards::Instance()->fetchValueDouble("Q2Min") && DISPred::ControlCards::Instance()->fetchValueDouble("YMax") > k[1] &&  DISPred::ControlCards::Instance()->fetchValueDouble("YMin") <k[1]  ){  
    if (!strcmp(DISPred::ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){  
    return jacob*DISPred::DISPredictor::Instance()->CalculateCCReducedCrossSection(k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]), k[0] )*DISPred::DISPredictor::Instance()->CalculateCCPropagator(k[0],k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) );
    } else {
      return jacob*DISPred::DISPredictor::Instance()->CalculateReducedCrossSection(k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]), k[0] )*DISPred::DISPredictor::Instance()->CalculatePropagator(k[0],k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) );
    }

  } else{
    return 0.0;
  }
}


double YgY (double *k, size_t dim, void *params)
{
  double jacob=k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]*k[1]); // multiply by (-q2)/(s*y**2)
  if (k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])<DISPred::ControlCards::Instance()->fetchValueDouble("XMax") && k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])>DISPred::ControlCards::Instance()->fetchValueDouble("XMin") && k[0]<DISPred::ControlCards::Instance()->fetchValueDouble("Q2Max") && k[0]>DISPred::ControlCards::Instance()->fetchValueDouble("Q2Min") && DISPred::ControlCards::Instance()->fetchValueDouble("YMax") > k[1] &&  DISPred::ControlCards::Instance()->fetchValueDouble("YMin") <k[1]  ){  
    if (!strcmp(DISPred::ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){  
    return k[1]*jacob*DISPred::DISPredictor::Instance()->CalculateCCReducedCrossSection(k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]), k[0] )*DISPred::DISPredictor::Instance()->CalculateCCPropagator(k[0],k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) );
    } else {
      return k[1]*jacob*DISPred::DISPredictor::Instance()->CalculateReducedCrossSection(k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]), k[0] )*DISPred::DISPredictor::Instance()->CalculatePropagator(k[0],k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) );
    }

  } else{
    return 0.0;
  }
}


double gQ2 (double *k, size_t dim, void *params)
{
  if (k[1]<DISPred::ControlCards::Instance()->fetchValueDouble("XMax") && k[1]>DISPred::ControlCards::Instance()->fetchValueDouble("XMin") && k[0]<DISPred::ControlCards::Instance()->fetchValueDouble("Q2Max") && k[0]>DISPred::ControlCards::Instance()->fetchValueDouble("Q2Min") && DISPred::ControlCards::Instance()->fetchValueDouble("YMax") > k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) &&  DISPred::ControlCards::Instance()->fetchValueDouble("YMin") <k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])  ){  
    if (!strcmp(DISPred::ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){  
    return k[0]*DISPred::DISPredictor::Instance()->CalculateCCReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculateCCPropagator(k[0],k[1] );
    } else {
      return k[0]*DISPred::DISPredictor::Instance()->CalculateReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculatePropagator(k[0],k[1] );
    }

  } else{
    return 0.0;
  }
}


double gX (double *k, size_t dim, void *params)
{
  if (k[1]<DISPred::ControlCards::Instance()->fetchValueDouble("XMax") && k[1]>DISPred::ControlCards::Instance()->fetchValueDouble("XMin") && k[0]<DISPred::ControlCards::Instance()->fetchValueDouble("Q2Max") && k[0]>DISPred::ControlCards::Instance()->fetchValueDouble("Q2Min") && DISPred::ControlCards::Instance()->fetchValueDouble("YMax") > k[0]/(DISPred::DISPredictor::Instance()->S()*k[1]) &&  DISPred::ControlCards::Instance()->fetchValueDouble("YMin") <k[0]/(DISPred::DISPredictor::Instance()->S()*k[1])  ){  
    if (!strcmp(DISPred::ControlCards::Instance()->fetchValueString("DISProcess").c_str(),"CC")){  
    return k[1]*DISPred::DISPredictor::Instance()->CalculateCCReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculateCCPropagator(k[0],k[1] );
    } else {
      return k[1]*DISPred::DISPredictor::Instance()->CalculateReducedCrossSection(k[1], k[0] )*DISPred::DISPredictor::Instance()->CalculatePropagator(k[0],k[1] );
    }

  } else{
    return 0.0;
  }
}






#endif

